'''
  Folding@Home Client Control (FAHControl)
  Copyright (C) 2010-2011 Stanford University

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import sys
import gtk
import traceback

from fah.util import parse_bool
from fah.util import status_to_color
from fah.util import get_widget_str_value
from fah.util import set_widget_str_value
from fah import SlotConfig


def get_option_mods(old_options, new_options):
    changes = {}
    
    # Deleted
    for name in old_options:
        if name not in new_options:
            changes[name + '!'] = None

    # Added and modified
    for name, value in new_options.items():
        if name not in old_options or value != old_options[name]:
            changes[name] = value

    return changes


class ClientConfig:
    base_uri = 'http://fah-web.stanford.edu/cgi-bin/main.py'
    user_uri = base_uri + '?qtype=userpage&username='
    queue_cols = ('id state statecolor percentdone percent').split()

    def __init__(self):
        self.last_updated = 0
        self.queue = []
        self.queue_map = {}
        self.slots = []
        self.options = {}
        self.core_options = {}
        self.info = ''
        self.log = []
        self.tooltip = ''


    def update_queue(self, queue):
        self.queue = queue
        self.queue_map = {}
        for values in self.queue:
            self.queue_map[values['id']] = values


    def update_user_info(self, app):
        # User
        user = self.options['user']
        app.user_info.set_label(user)
        app.user_info.set_uri(self.user_uri + user)

        # Team
        team = self.options['team']
        app.team_info.set_label(team)
        iter = app.team_stats_link.get_active_iter()
        uri = app.team_stats_links_list.get(iter, 1)[0]
        app.team_info.set_uri(uri % {'team': team})


    def reset_user_info(self, app):
        # User
        app.user_info.set_label('')
        app.user_info.set_uri('')

        # Team
        app.team_info.set_label('')
        app.team_info.set_uri('')


    def get_selected_queue_entry(self, app):
        selection = app.queue_tree.get_selection()
        if selection is None: return None
        iter = selection.get_selected()[1]
        if iter: return app.queue_list.get_value(iter, 1) # Queue ID
        else: return None


    def get_selected_slot(self, app):
        selection = app.slot_status_tree.get_selection()
        if selection is None: return None
        iter = selection.get_selected()[1]
        if iter:
            id = int(app.slot_status_list.get_value(iter, 0)) # Slot ID
            for slot in self.slots:
                if slot.id == id: return slot


    def update_queue_ui(self, app):
        if not self.queue:
            app.queue_list.clear()
            return

        # Get selection
        selected = self.get_selected_queue_entry(app)

        # Reload queue list
        app.queue_list.clear()
        selected_row = 0
        for values in self.queue:
            unit_id = values['unit']
            queue_id = values['id']
            status = values['state'].title()
            color = status_to_color(status)
            progress = values['percentdone']
            percent = float(progress[:-1])
            eta = values['eta']
            if eta == '0.00 secs': eta = 'Unknown'
            credit = values['creditestimate']
            if float(credit) == 0: credit = 'Unknown'

            prcg = '%s (%s, %s, %s)' % (values['project'], values['run'],
                                        values['clone'], values['gen'])
            iter = app.queue_list.append([unit_id, queue_id, status, color,
                                          progress, percent, eta, credit, prcg])

            if queue_id == selected: selected_row = iter

        # Selected the first item if nothing is selected
        if not selected or not selected in self.queue_map:
            selected = self.queue_map.keys()[0]
            selected_row = app.queue_list.get_iter_first()

        # Restore selection
        app.queue_tree.get_selection().select_iter(selected_row)


    def update_work_unit_info(self, app):
        if not self.queue:
            self.reset_work_unit_info(app)
            return

        # Get selected queue entry
        selected = self.get_selected_queue_entry(app)
        if selected is None: return
        entry = self.queue_map[selected]

        # Load info
        for name, value in entry.items():
            if name in app.queue_widgets:
                if (name in ['basecredit', 'creditestimate', 'ppd'] and \
                        float(value) == 0) or value == '<invalid>' or \
                        value == '0.00 secs': value = 'Unknown'

                widget = app.queue_widgets[name]
                set_widget_str_value(widget, value)

        # Status
        status = entry['state'].title()
        color = status_to_color(status)
        widget = app.queue_widgets['state']
        if color: widget.set_markup('<span background="%s">%s</span>' %
                                    (color, status))
        else: widget.set_markup(status)

        # PRCG
        prcg = '%s (%s, %s, %s)' % (
            entry['project'], entry['run'], entry['clone'], entry['gen'])
        set_widget_str_value(app.queue_widgets['prcg'], prcg)
        
        # Links
        for name in ['cs', 'ws', 'project']:
            widget = app.queue_widgets[name]
            value = str(entry[name])

            if name in ['cs', 'ws']:
                uri = 'http://fah-web.stanford.edu/logs/%s.log.html' % value
            else:
                uri = 'http://fah-web.stanford.edu/cgi-bin/fahproject.' \
                    'overusingIPswillbebanned?p=' + value

            widget.set_uri(uri)


    def select_slot(self, app):
        # Get selected slot
        slot = self.get_selected_slot(app)
        if slot is None: return

        # Check status
        if not slot.status.upper() in ['RUNNING', 'FINISHING']: return

        # Get associated queue ID
        for entry in self.queue:
            if int(entry['slot']) == slot.id and \
                    entry['state'].upper() in ['RUNNING', 'FINISHING']:
                # Find unit_id in the queue list entry and select row
                unit_id = entry['unit']

                list = app.queue_list
                iter = list.get_iter_first()
                while iter is not None:
                    if list.get_value(iter, 0) == unit_id:
                        app.queue_tree.get_selection().select_iter(iter)
                        break
                    iter = list.iter_next(iter)
                
                # Update the UI
                self.update_work_unit_info(app)

                break


    def select_queue_slot(self, app):
        # Get unit ID of selected queue entry
        selected = self.get_selected_queue_entry(app)
        if selected is None: return

         # Get associated slot ID
        entry = self.queue_map[selected]
        slot = int(entry['slot'])

        # Find and select the slot
        list = app.slot_status_list
        iter = list.get_iter_first()
        while iter is not None:
            if int(list.get_value(iter, 0)) == slot:
                app.slot_status_tree.get_selection().select_iter(iter)
                break
            iter = list.iter_next(iter)

        # Update the UI
        self.update_work_unit_info(app)


    def reset_work_unit_info(self, app):
        for widget in app.queue_widgets.values():
            set_widget_str_value(widget, None)


    def update_info(self, app):
        port = app.info
        
        # Clear
        for child in port.get_children(): port.remove(child)

        # Alignment
        align = gtk.Alignment(0, 0, 1, 1)
        align.set_padding(4, 4, 4, 4)
        port.add(align)

        # Vertical box
        vbox = gtk.VBox()
        align.add(vbox)

        for category in self.info:
            name = category[0]
            category = category[1:]
            
            # Frame
            frame = gtk.Frame('<b>%s</b>' % name)
            frame.set_shadow_type(gtk.SHADOW_ETCHED_IN)
            frame.get_label_widget().set_use_markup(True)
            vbox.pack_start(frame, False)

            if not len(category): continue

            # Alignment
            align = gtk.Alignment(0, 0, 1, 1)
            align.set_padding(0, 0, 12, 0)
            frame.add(align)

            # Table
            table = gtk.Table(len(category), 2)
            table.set_col_spacing(0, 5)
            align.add(table)

            row = 0
            for name, value in category:
                # Name
                label = gtk.Label('<b>%s</b>' % name)
                label.set_use_markup(True)
                label.set_alignment(1, 0.5)
                table.attach(label, 0, 1, row, row + 1, gtk.FILL, gtk.FILL)

                # Value
                if value.startswith('http://'):
                    label = gtk.LinkButton(value, value)
                    label.set_relief(gtk.RELIEF_NONE)
                    label.set_property('can-focus', False)

                else: label = gtk.Label(value)

                label.set_alignment(0, 0.5)
                label.modify_font(app.mono_font)
                table.attach(label, 1, 2, row, row + 1, yoptions = gtk.FILL)

                row += 1

        port.show_all()


    def update_options(self, app):
        used = set()

        for name, widget in app.client_option_widgets.items():
            name = name.replace('_', '-')
            used.add(name)

            try:
                set_widget_str_value(widget, self.options[name])

            except: # Don't let one bad widget kill everything
                print 'WARNING: failed to set widget "%s"' % name

        # Setup passkey and password entries
        app.passkey_validator.set_good()
        app.password_validator.set_good()

        # Set core priority radio button
        if self.options['core-priority'] == 'idle':
            app.client_option_widgets['core_priority'].set_active(True)
        else: app.core_priority_low.set_active(True)

        # Extra core options
        app.core_option_list.clear()
        if 'extra-core-args' in self.options:
            used.add('extra-core-args')

            args = self.options['extra-core-args'].split()
            for arg in args: app.core_option_list.append([arg])

        # Remaining options
        app.option_list.clear()
        for name, value in self.options.items():
            if name not in used:
                app.option_list.append([name, value])


    def update_status_slots(self, app):
        # Preserve selection
        selection = app.slot_status_tree.get_selection()
        model, selected = selection.get_selected_rows()

        # Update list
        app.slot_status_list.clear()
        for slot in self.slots:
            id = slot.id
            status = slot.status.title()
            color = status_to_color(status)
            description = slot.description.replace('"', '')
            app.slot_status_list.append((id, status, color, description))

        # Restore selection
        for path in selected:
            selection.select_path(path)


    def update_slots_ui(self, app):
        app.slot_list.clear()
        for slot in self.slots:
            slot.add_to_ui(app)


    def scroll_log_to_end(self, app):
        mark = app.log.get_mark('end')
        app.log.move_mark(mark, app.log.get_end_iter())
        app.log_view.scroll_mark_onscreen(mark)


    def log_clear(self, app):
        app.log.set_text('')
        self.log = []
        

    def log_add(self, app, text):
        app.log.insert(app.log.get_end_iter(), text)
        self.log.append(text)        


    def update_log(self, app):
        app.log.set_text('')
        for text in self.log:
            app.log.insert(app.log.get_end_iter(), text)

        self.scroll_log_to_end(app)


    def update_status_ui(self, app):
        self.update_queue_ui(app)
        self.update_status_slots(app)
        self.update_work_unit_info(app)
        self.update_systray(app)


    def reset_status_ui(self, app):
        self.reset_work_unit_info(app)
        app.queue_list.clear()
        app.slot_status_list.clear()
        app.log.set_text('')
        self.reset_systray(app)


    def get_running(self):
        for unit in self.queue:
            if unit['state'].upper() == 'RUNNING': return True
        return False
        

    def update_systray(self, app):
        if not app.tray_icon: return

        # Only novice mode on Windows
        basic_mode = app.get_mode() == 'novice' or sys.platform == 'win32'

        # Team
        s = '\nTeam: ' + self.options['team']

        # Status
        s += '\nStatus: '
        running = self.get_running()
        if running: s += 'Running'
        else: s += 'Inactive'

        # Unit progress
        if running and not basic_mode:
            s += '\nProgress: '
            for unit in self.queue:
                if unit['state'].upper() == 'RUNNING':
                    s += '\n  Unit %s: %s' % (unit['id'], unit['percentdone'])
                    
        # Name
        name = 'Name: ' + self.options['user']

        # Windows restricts the sys-tray tooltip to 63 or 127 characters
        if sys.platform == 'win32':
            l = len(s) + len(name) - 63
            if 0 < l: name = name[:-(l + 3)] + '...'

        s = name + s

        if self.tooltip != s:
            app.tray_icon.set_tooltip(s)
            self.tooltip = s


    def reset_systray(self, app):
        if not app.tray_icon: return
        app.tray_icon.set_tooltip('Inactive')


    def update_ui(self, app):
        # Update everything
        self.update_user_info(app)
        self.update_info(app)
        self.update_options(app)
        self.update_slots_ui(app)
        self.update_log(app)
        self.update_status_ui(app)


    def get_option_changes(self, app):
        used = set()
        options = {}

        # Core priority radio button
        used.add('core-priority')
        if app.client_option_widgets['core_priority'].get_active():
            if self.options['core-priority'] != 'idle':
                options['core-priority'] = 'idle'
        elif self.options['core-priority'] == 'idle':
            options['core-priority'] = 'low'

        # Extra core options
        used.add('extra-core-args')
        if 'extra-core-args' in self.options:
            old_args = self.options['extra-core-args'].split()
        else: old_args = []

        new_args = []
        def add_arg(model, path, iter, data):
            new_args.append(model.get(iter, 0)[0])
        app.core_option_list.foreach(add_arg, None)

        if old_args != new_args:
            if new_args: options['extra-core-args'] = ' '.join(new_args)
            else: options['extra-core-args!'] = None

        # Extra options
        def check_option(model, path, iter, data):
            name, value = model.get(iter, 0, 1)
            used.add(name)
            if not name in self.options or self.options[name] != value:
                options[name] = value

        app.option_list.foreach(check_option, None)
        
        # Main options
        for name, widget in app.client_option_widgets.items():
            name = name.replace('_', '-')
            if name in used: continue
            value = self.options[name]
            used.add(name)

            try:
                value = get_widget_str_value(widget)
                if value is None: options[name + '!'] = None
                elif value != self.options[name]: options[name] = value

            except Exception, e: # Don't let one bad widget kill everything
                print 'WARNING: failed to save widget "%s": %s' % (name, e)

        # Removed options
        for name in self.options:
            if not name in used:
                options[name + '!'] = None

        return options

    
    def get_slot_changes(self, app):
        # Get new slots
        new_slots = []
        def add_slot(model, path, iter, data = None):
            new_slots.append(model.get(iter, 2)[0].slot)
        app.slot_list.foreach(add_slot)

        # Get old slot IDs
        old_slot_map = {}
        for slot in self.slots: old_slot_map[slot.id] = slot

        # Get new slot IDs
        new_slot_ids = set()
        for slot in new_slots: new_slot_ids.add(slot.id)

        # Find deleted slot IDs
        deleted = []
        for id in old_slot_map:
            if id not in new_slot_ids: deleted.append(id)

        # Find added and modified slots
        added = []
        modified = []
        for slot in new_slots:
            # Added
            if slot.id == -1: added.append((slot.type, slot.options))
            else:
                old_slot = old_slot_map[slot.id]
                options = get_option_mods(old_slot.options, slot.options)
                if options or old_slot.type != slot.type:
                    modified.append((slot.id, slot.type, options))

        return (deleted, added, modified)


    def get_changes(self, app):
        return self.get_option_changes(app), self.get_slot_changes(app)


    def has_changes(self, app):
        options, slots = self.get_changes(app)
        return options or slots != ([], [], [])
