'''
  Folding@Home Client Control (FAHControl)
  Copyright (C) 2010-2011 Stanford University

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import sys
import os
import re
import traceback
import platform

import gtk
import glib
import pygtk
pygtk.require("2.0")
import pango
import webbrowser
import shlex
import subprocess

# OSX integration
if sys.platform == 'darwin':
    from gtk_osxapplication import *

from fah import *
from fah.db import *
from fah.util import *


def set_tree_view_font(widget, font):
    for widget in iterate_container(widget):
        if isinstance(widget, gtk.TreeView):
            widget.modify_font(font)


def append_tree_entry(model, path, iter, selection):
    selection.append((path, iter))


def get_tree_selection(tree_view):
    selection = []
    tree_view.get_selection().selected_foreach(append_tree_entry, selection)
    return selection


def remove_tree_selection(tree_view, callback = None):
    for path, iter in get_tree_selection(tree_view):
        if callback is not None: callback(path, iter)
        tree_view.get_model().remove(iter)


def add_osx_menu_item(app, group, name, callback):
    item = gtk.MenuItem(name)
    item.show()
    item.connect('activate', callback)
    app.add_app_menu_item(group, item)


def load_fahcontrol_db():
    db = Database(os.path.join(get_home_dir(), 'FAHControl.db'))
    db.validate()

    return db
    

class FAHControl(SingleAppServer):
    client_cols = 'name status status_color address'.split()

    team_stats_links = [
        ['Folding@home', 'http://fah-web.stanford.edu/cgi-bin/main.py'
         '?qtype=teampage&teamnum=%(team)s'],
        ['Extreme Overclocking', 'http://folding.extremeoverclocking.com/'
         'team_summary.php?t=%(team)s'],
        ['Kakao Stats', 'http://kakaostats.com/tsum.php?t=%(team)s'],
        ['[H]ard Folding', 'http://www.hardfolding.com/fh_stats/index.php'
         '?pz=101&tnum=%(team)s'],
        ]


    def __init__(self, glade = 'FAHControl.glade'):
        SingleAppServer.__init__(self)

        # Windows needs this but OSX accelerators crash with this
        if sys.platform != 'darwin': gtk.gdk.threads_init()

        # Vars
        self.clients = {}
        self.clientsByAddress = {}
        self.active_client = None
        self.client_is_online = False
        self.selected_clients = set()
        self.status_clear_time = None
        self.mouse_in_tray_menu = None
        self.window_visible = False
        self.viewer = None
        self.last_db_flush = 0
        self.last_clients_update = 0
        self.error_dialog = None
        self.restore_dialogs = []

        # Open database
        try:
            self.db = load_fahcontrol_db()

        except Exception, e:
            print e
            sys.exit(1)

        # OSX integration
        if sys.platform == 'darwin':
            self.osx_app = OSXApplication()
            self.osx_app.set_dock_icon_pixbuf(get_icon('large'))
            self.osx_app.set_use_quartz_accelerators(True)

        # URI hook
        gtk.link_button_set_uri_hook(self.on_uri_hook, None)

        # Style
        rcfile = os.path.join(os.path.expanduser("~"), '.FAHClient.gtkrc')
        if os.path.exists(rcfile): gtk.rc_parse(rcfile)
        #gtk.rc_parse('/usr/share/themes/AbsE/gtk-2.0/gtkrc')
        #gtk.rc_parse('themes/TheGreyHero.gtkrc')
        self.mono_font = pango.FontDescription('Monospace 10')

        # Default icon
        gtk.window_set_default_icon(get_icon('small'))

        # Build GUI
        self.builder = builder = gtk.Builder()
        try:
            if len(glade) < 1024: builder.add_from_file(glade)
            else: builder.add_from_string(glade)
        except:
            self.error('Failed to load UI file: %s' % glade)
            sys.exit(1)

        # Main window
        self.window = builder.get_object('window')
        self.window.set_geometry_hints(None, 440, 256, -1, -1, 800, 512)
        set_tree_view_font(self.window, self.mono_font)
        self.status_bar = builder.get_object('status_bar')

        # Panes
        self.panes = WidgetMap(self.window, 'paned')
        for name, pane in self.panes.items():
            prop = name + '_position'

            # Load current value
            if self.db.has(prop):
                pane.set_property('position', int(self.db.get(prop)))

            pane.connect('notify::position', self.store_property, prop)

        # Tray
        if sys.platform != 'darwin':
            self.tray_icon = builder.get_object('tray_icon')
            self.tray_icon.set_from_pixbuf(get_icon('small'))
        else: self.tray_icon = None
        self.tray_menu = builder.get_object('tray_menu')
        self.restore_tray_item = builder.get_object('restore_tray_item')
        item = builder.get_object('viewer_tray_item')
        item.set_image(get_viewer_icon_image('tiny'))

        # Dialogs
        self.preferences_dialog = builder.get_object('preferences_dialog')
        self.client_dialog = builder.get_object('client_dialog')
        self.options_dialog = builder.get_object('options_dialog')
        self.core_options_dialog = builder.get_object('core_options_dialog')
        self.slot_dialog = builder.get_object('slot_dialog')
        self.about_dialog = builder.get_object('about_dialog')
        # Note: The order of these dialogs is important since they are restored
        #   in this order.  Child dialogs cannot be restored before their
        #   parents so they must be last.  See minimize() and restore() below.
        self.dialogs = [
            self.about_dialog, self.preferences_dialog, self.client_dialog, 
            self.slot_dialog, self.options_dialog, self.core_options_dialog]

        # Dialog & window sizes
        self.windows = {
            'preferences': self.preferences_dialog,
            'client': self.client_dialog,
            'options': self.options_dialog,
            'core_options': self.core_options_dialog,
            'slot': self.slot_dialog,
            'about': self.about_dialog,
            'main': self.window,
            }
        for name, win in self.windows.items():
            if self.db.has(name + '_width'):
                width = int(self.db.get(name + '_width'))
            else: width = -1

            if self.db.has(name + '_height'):
                height = int(self.db.get(name + '_height'))
            else: height = -1
            
            if 0 <= width and 0 <= height: win.resize(width, height)

            win.connect('configure_event', self.store_dimensions, name)

        # Tool bar
        viewer_button = builder.get_object('viewer_button')
        viewer_button.set_icon_widget(get_viewer_icon_image('tiny'))

        self.mode_combo =  builder.get_object('mode_combo')

        # About Dialog
        icon = builder.get_object('about_icon')
        icon.set_from_pixbuf(get_icon('medium'))
        about_version = builder.get_object('about_version')
        about_version.set_markup('<b>Version: %s</b>' % version)

        # Preferences
        self.theme_list = builder.get_object('theme_list')
        self.load_themes(self.theme_list)

        # Client list
        self.client_hpane = builder.get_object('client_hpane')
        self.client_notebook = builder.get_object('client_notebook')
        self.client_config_notebook = \
            builder.get_object('client_config_notebook')
        self.client_tree = builder.get_object('client_tree_view')
        self.client_list = builder.get_object('client_list')
        self.client_label = builder.get_object('client_label')
        self.client_config_label = builder.get_object('client_config_label')
        self.client_tree.grab_focus()
        selection = self.client_tree.get_selection()
        selection.connect('changed', self.on_client_selection_changed)

        # Option lists
        self.option_tree, self.option_list = self.connect_option_view('')
        self.slot_option_tree, self.slot_option_list =\
            self.connect_option_view('slot_')
        self.core_option_tree = builder.get_object('core_option_tree_view')
        self.core_option_list = builder.get_object('core_option_list')
        self.option_tree.get_selection().set_mode(gtk.SELECTION_MULTIPLE)
        self.slot_option_tree.get_selection().set_mode(gtk.SELECTION_MULTIPLE)
        self.core_option_tree.get_selection().set_mode(gtk.SELECTION_MULTIPLE)

        # Option dialog
        self.option_name_entry = builder.get_object('option_name_entry')
        self.option_value_entry = builder.get_object('option_value_entry')

        # Core option dialog
        self.core_option_entry = builder.get_object('core_option_entry')

        # User info
        self.user_info = builder.get_object('user_info')
        self.team_info = builder.get_object('team_info')
        #self.user_stats_link = builder.get_object('user_stats_link')
        self.team_stats_link = builder.get_object('advanced_team_stats_link')
        self.team_stats_links_list = \
            builder.get_object('team_stats_links_list')
        map(self.team_stats_links_list.append, self.team_stats_links)
        self.team_stats_link.set_active(0)

        # Client config
        self.local_client_frame = builder.get_object('local_client_frame')
        self.local_client_button = builder.get_object('local_client_button')
        self.core_priority_low = builder.get_object('core_priority_low')

        # Slot lists
        self.slot_status_tree = builder.get_object('slot_status_tree_view')
        self.slot_status_tree.get_selection().set_mode(gtk.SELECTION_SINGLE)
        self.slot_status_list = builder.get_object('slot_status_list')
        self.slot_tree = builder.get_object('slot_tree_view')
        self.slot_list = builder.get_object('slot_list')
        self.slot_menu = builder.get_object('slot_menu')
        view_slot_item = builder.get_object('view_slot_item')
        view_slot_item.get_image().set_from_pixbuf(get_viewer_icon('tiny'))

        # Slot dialog
        self.slot_type_uni = builder.get_object('slot_type_uni')
        self.slot_type_smp = builder.get_object('slot_type_smp')
        self.slot_type_gpu = builder.get_object('slot_type_gpu')

        # Queue list
        self.queue_tree = builder.get_object('queue_tree')
        self.queue_list = builder.get_object('queue_list')

        # Info
        self.info = builder.get_object('info_view_port')

        # Log
        self.log_view = builder.get_object('log_text_view')
        self.log_view.modify_font(self.mono_font)
        self.log = builder.get_object('log_buffer')
        self.log.create_mark('end', self.log.get_end_iter())

        # OSX integration
        if sys.platform == 'darwin':
            # Conflicts with the standard OSX interface
            builder.get_object('hide_button').hide()

            # Setup dock menu
            self.osx_menu = builder.get_object('osx_tray_menu')
            self.osx_app.set_dock_menu(self.osx_menu)
            self.osx_menu_widgets = WidgetMap(self.osx_menu, prefix = 'osx_')

            # Create application menu
            self.osx_menubar = gtk.MenuBar()
            self.osx_menubar.show_all()
            self.osx_app.set_menu_bar(self.osx_menubar)
            self.osx_group = self.osx_app.add_app_menu_group()
            self.osx_menu.foreach(self.osx_add_to_menu)


        # Widget maps
        self.client_entries = WidgetMap(self.client_dialog, '_entry')
        self.client_option_widgets = \
            WidgetMap(self.client_config_notebook, '_option')
        self.client_config_tabs = WidgetMap(self.client_config_notebook, '_tab')
        self.slot_option_widgets = WidgetMap(self.slot_dialog, '_option')
        roots = [self.window, self.client_dialog]
        self.advanced_widgets = WidgetMap(roots, None, 'advanced_')
        self.expert_widgets = WidgetMap(roots, None, 'expert_')
        self.preference_widgets = WidgetMap(self.preferences_dialog, '_pref')
        widget = builder.get_object('advanced_unit_frame')
        self.queue_widgets = WidgetMap(widget, None, 'queue_')

        # Validators
        EntryValidator(self, self.client_option_widgets['user'], r'^\S+$',
                       'User name must be a non-empty string containing no '
                       'alphanumeric characters, standard punctuation and '
                       'no white-space.')

        self.passkey_validator = \
            PasswordValidator(self, builder.get_object('passkey_option'),
                              builder.get_object('passkey_reenter'),
                              builder.get_object('passkey_valid_image'),
                              builder.get_object('passkey_valid_text'),
                              r'^[0-9a-fA-F]{0,32}$', 'The passkey must be a '
                              '32 character hexadecimal string.')

        self.password_validator = \
            PasswordValidator(self, builder.get_object('password_option'),
                              builder.get_object('password_reenter'),
                              builder.get_object('password_valid_image'),
                              builder.get_object('password_valid_text'))

        # Fix client port default
        port = builder.get_object('port_adjustment')
        port.set_value(36330)

        # Fix text-view labels
        self.window.realize()
        style = self.window.get_style()
        color = style.bg[gtk.STATE_NORMAL]
        for id in range(1, 300):
            label = builder.get_object('textview%d' % id)
            if label is not None:
                label.modify_base(gtk.STATE_INSENSITIVE, color)

        # Connect signals
        builder.connect_signals(self)
        self.builder = builder = None # Discard builder

        # Load
        self.load_clients()
        self.preferences_load()

        # If we don't have any clients add the default client
        if not len(self.clients):
            client = Client(self, 'local', '127.0.0.1', 36330, '')
            self.add_client(client)
            self.save_clients()

        # Select first client
        self.select_first_client()

        # Start client notebook deactivated
        self.deactivate_client()


    # Main loop
    def run(self):
        self.check_clients() # Slightly faster load?

        if self.tray_icon: self.tray_icon.set_visible(True)
        self.timer_id = gobject.timeout_add(100, self.on_timer)

        if sys.platform == 'darwin': self.osx_app.ready()

        # Windows needs this but OSX accelerators crash with this
        if sys.platform != 'darwin': gtk.gdk.threads_enter()
        gtk.main()
        if sys.platform != 'darwin': gtk.gdk.threads_leave()

        try:
            self.db.flush_queued()
        except Exception, e: print e

        if self.tray_icon: self.tray_icon.set_visible(False)


    # Util
    def osx_add_to_menu(self, widget):
        if isinstance(widget, gtk.SeparatorMenuItem):
            self.osx_group = self.osx_app.add_app_menu_group()

        elif isinstance(widget, gtk.MenuItem):
            name = widget.child.get_text()

            def activate_item(widget, target):
                target.emit('activate')

            item = gtk.MenuItem(name)
            item.show()
            item.connect('activate', activate_item, widget)
            self.osx_app.add_app_menu_item(self.osx_group, item)
        

    def connect_option_cell(self, name, model, col):
        cell = self.builder.get_object(name)
        cell.connect('edited', self.on_option_edit, model, col)


    def connect_option_view(self, prefix):
        tree = self.builder.get_object(prefix + 'option_tree_view')
        model = self.builder.get_object(prefix + 'option_list')

        self.connect_option_cell(prefix + 'option_name_cell', model, 0)
        self.connect_option_cell(prefix + 'option_value_cell', model, 1)

        return tree, model

    # Timer functions
    def check_clients(self):
        # Make sure there is a selected client
        if not len(self.selected_clients): self.select_first_client()

        # Start local client
        if self.autostart_client: self.start_local_client()

        # Update clients
        for client in self.clients.values(): client.update(self)

        # (De)activate client
        if self.active_client:
            if not self.active_client.is_updated():
                self.deactivate_client()

        else: self.activate_client() # Try to activate

        # Check if active and online
        if self.active_client and self.active_client.is_online():
            self.client_notebook.set_sensitive(True)
        else: self.client_notebook.set_sensitive(False)

        # Update status bar
        if self.status_clear_time and self.status_clear_time < time.time():
            self.status_bar.pop(0)
            self.status_clear_time = None


    def check_hide_popup_menu(self, data = None):
        if sys.platform not in ['win32', 'darwin'] : return
        if self.mouse_in_tray_menu and self.mouse_in_tray_menu < time.time():
            self.mouse_in_tray_menu = None
            self.tray_menu.hide()
            self.tray_menu.get_window().hide()


    def on_timer(self):
        try:
            self.check_hide_popup_menu()
            self.check_clients()
            self.viewer_check()

            if self.ping.isSet():
                self.ping.clear()
                self.restore()

            if 2.5 < time.time() - self.last_db_flush:
                self.last_db_flush = time.time()
                self.db.flush_queued()

        except:
            traceback.print_exc()

        return True # Keep running


    # Actions
    def quit(self):
        gtk.main_quit()

        self.viewer_close()

        # Necessary to make sure we shutdown in Windows
        self.tray_menu.hide()
        glib.source_remove(self.timer_id)

        for client in self.clients.values(): client.close()
        self.shutdown() # Shutdown SingleAppServer


    def set_status(self, text):
        self.status_bar.pop(0)
        self.status_bar.push(0, text)
        self.status_clear_time = time.time() + 10


    def set_mode(self, mode):
        mode = mode.lower()

        if mode == 'novice' and 1 < len(self.clients):
            self.error('Novice mode is only valid with one client')
            mode = 'advanced'
            set_widget_str_value(self.mode_combo, mode)

        show_advanced = mode == 'advanced' or mode == 'expert'
        show_expert = mode == 'expert'

        for w in self.advanced_widgets.values():
            w.set_property('visible', show_advanced)

        for w in self.expert_widgets.values():
            w.set_property('visible', show_expert)

        return mode


    # Preference methods
    def load_theme(self, theme):
        path = gtk.rc_get_theme_dir()
        rc = path + '/' + theme + '/gtk-2.0/gtkrc'

        gtk.rc_parse(rc) # This can cause pygtk to lock up
        screen = self.window.get_screen()
        gtk.rc_reset_styles(gtk.settings_get_for_screen(screen))


    def load_themes(self, list):
        path = gtk.rc_get_theme_dir()
        if not os.path.exists(path): return
        for name in os.listdir(path):
            rc = path + '/' + name + '/gtk-2.0/gtkrc'
            if os.path.exists(rc): list.append([name, rc])


    def get_mode(self):
        return self.mode_combo.get_active_text().lower()


    def get_pref(self, name):
        widget = self.preference_widgets[name]
        return get_widget_str_value(widget)


    def get_viz_render_mode(self):
        value = self.get_pref('viz_render_mode')
        
        if value == 'Space Filling': return 1
        if value == 'Ball And Stick': return 2
        if value == 'Stick': return 3
        if value == 'Cartoon Space Filling': return 4
        if value == 'Cartoon Ball And Stick': return 5

        return 1


    def preferences_set(self):
        self.set_mode(self.get_mode())
        self.autostart_client = parse_bool(self.get_pref('autostart_client'))
        self.quit_on_close = parse_bool(self.get_pref('quit_on_close'))


    def preferences_load(self):
        minimized = False

        # Mode
        if self.db.has('mode'): mode = self.db.get('mode')
        else: mode = 'novice'
        self.set_mode(mode)
        set_widget_str_value(self.mode_combo, mode)

        # Preferences dialog
        for name, widget in self.preference_widgets.items():
            value = None

            if self.db.has(name):
                value = self.db.get(name)
                if name == 'minimized': minimized = parse_bool(value)
                elif name == 'autostart_client':
                    self.autostart_client = parse_bool(value)
                elif name == 'quit_on_close':
                    self.quit_on_close = parse_bool(value)
                elif name == 'theme': self.load_theme(value)

            elif name == 'minimized': value = str(minimized)
            elif name == 'autostart_client':
                value = 'True'
                self.autostart_client = True
            elif name == 'client_command': value = 'FAHClient'
            elif name == 'quit_on_close':
                value = 'False'
                self.quit_on_close = False
            elif name == 'theme': value = None
            elif name == 'viz_command': value = 'FAHViewer'
            elif name == 'viz_fullscreen': value = 'False'
            elif name == 'viz_width': value = '800'
            elif name == 'viz_height': value = '600'
            elif name == 'viz_render_mode': value = 'Space Filling'
            elif name == 'viz_cycle_snapshots': value = 'True'
            else: raise Exception, 'Unknown preference widget "%s"' % name

            if value is not None: set_widget_str_value(widget, value)

        # Minimized
        if minimized: self.minimize()
        else: self.restore()


    def preferences_dialog_init(self):
        for name, widget in self.preference_widgets.items():
            if self.db.has(name):
                value = self.db.get(name)
                set_widget_str_value(widget, value)


    def preferences_save(self):
        for name, widget in self.preference_widgets.items():
            value = get_widget_str_value(widget)
            if value is None: self.db.clear(name, False)
            else: self.db.set(name, value, False)

        self.db.commit()


    # Client methods
    def select_first_client(self):
        iter = self.client_list.get_iter_first()
        if iter: self.client_tree.get_selection().select_iter(iter)


    def start_local_client(self):
        for client in self.clients.values():
            if client.is_local():
                client.autostart(self)
                break


    def activate_client(self):
        if self.active_client: return
        if not len(self.selected_clients): return

        # Check that all selected clients are active
        for client in self.selected_clients:
            if not client.is_updated(): return

        # Activate client(s)
        for client in self.selected_clients:
            self.active_client = client
            self.active_client.update_status_ui(self)
            break # TODO only supporting one active client right now

        self.last_clients_update = 0
        self.update_client_status()


    def deactivate_client(self):
        if self.active_client:
            self.active_client.reset_status_ui(self)
            self.active_client = None

        self.client_label.set_markup('<b>Client: inactive</b>')
        self.update_client_status()


    def update_client_status(self):
        if len(self.selected_clients):
            client = list(self.selected_clients)[0]

            text = '<b>Client: %s' % client.name
            status = client.get_status()
            color = status_to_color(status)
            text += ' <span background="%s">%s</span></b>' % (color, status)

            if self.active_client and self.active_client.config.get_running():
                text += ' Running'
            else: text += ' Inactive'

        else: text = '<b>Client: no client selected</b>'
            
        self.client_label.set_markup(text)
        self.client_config_label.set_markup(text)

        if not len(self.selected_clients) and self.tray_icon:
            self.tray_icon.set_tooltip('Inactive')


    def save_clients(self):
        self.db.delete('clients')

        for client in self.clients.values():
            client.save(self.db)

        self.db.commit()


    def load_clients(self):
        for row in self.db.select('clients'):
            client = Client(self, row['name'], row['address'], int(row['port']),
                            row['password'])
            self.add_client(client)


    def clear_clients(self):
        for client in self.clients: client.close()
        self.clients.clear()
        self.clientsByName.clear()
        self.client_list.clear()


    def save_client_config(self, client):
        try:
            options, slots = client.config.get_changes(self)
            if not (options or slots != ([], [], [])): return True

            # Validate passkey
            if not self.passkey_validator.is_good():
                raise Exception, 'Passkey is invalid'

            # Validate password
            if not self.password_validator.is_good():
                raise Exception, 'Client password is invalid'

            self.deactivate_client()

            self.set_status('Saving...')

            save = False
            for client in self.selected_clients:
                # TODO check returned error count
                client.save_config(options, slots)
                client.update(self) # Slightly faster save

                # Update password if changed on client
                if 'password' in options:
                    client.set_password(options['password'])
                    save = True

            if save: self.save_clients() # Password was updated

            return True

        except Exception, msg:
            self.set_status('Save Failed')
            self.error(msg)
            return False


    def check_duplicate_client_name(self, name):
        if name in self.clients:
            self.error('Client with name "%s" is already in client list' % name)
            return True
        return False


    def check_duplicate_client_address(self, address):
        if address in self.clientsByAddress:
            self.error('Client with address "%s" is already in client list' %
                       address)
            return True
        return False


    def update_client(self, client, name, address, port, password):
        reload = False
        old_name = client.name

       # Check for duplicates
        if client.name != name:
            if self.check_duplicate_client_name(name): return False

        new_address = Client.make_address(address, port)
        if client.get_address() != new_address:
            if self.check_duplicate_client_address(new_address): return False

        # Update
        if client.name != name:
            del self.clients[client.name]
            client.name = name
            self.clients[name] = client

        if client.get_address() != new_address:
            del self.clientsByAddress[client.get_address()]
            client.set_address(address, port)
            self.clientsByAddress[new_address] = client
            reload = True

        if client.get_password() != password:
            client.set_password(password)
            reload = True

        # Update client row
        row = client.get_row(self)

        # Find client in client_list
        iter = self.client_list.get_iter_first()
        while iter is not None:
            if old_name == self.client_list.get_value(iter, 0):
                path = self.client_list.get_path(iter)

                for i in range(len(row)):
                    self.client_list.set(iter, i, row[i])
                self.client_list.row_changed(path, iter)

                break

            iter = self.client_list.iter_next(iter)

        # Reload
        if reload:
            client.reconnect()
            self.deactivate_client()

        return True


    def add_client(self, client):
        name = client.name
        address = client.get_address()

        # Check for duplicates
        if self.check_duplicate_client_name(name): return False
        if self.check_duplicate_client_address(address): return False

        # Add it
        self.clients[name] = client
        self.clientsByAddress[address] = client
        self.client_list.append(client.get_row(self))

        return True


    def remove_client(self, client):
        client.close()
        del self.clients[client.name]
        del self.clientsByAddress[client.get_address()]
        if client is self.active_client: self.active_client = None


    def remove_client_path(self, path, iter):
        name = self.client_list.get(iter, 0)[0]
        self.remove_client(self.clients[name])


    def get_selected_clients(self):
        selection = get_tree_selection(self.client_tree)
        names = map(lambda item: self.client_list.get(item[1], 0)[0], selection)
        return set(map(self.clients.get, names))


    def edit_client(self, client):
        self.client_dialog.client = client
        client.load_dialog(self)

        if self.active_client and self.active_client.is_online():
            for name, widget in self.client_config_tabs.items(): widget.show()
            self.client_dialog.config_hidden = False
            self.set_mode(self.get_mode())

        else:
            for name, widget in self.client_config_tabs.items():
                if name != 'connection': widget.hide()
            self.client_dialog.config_hidden = True

        self.client_dialog.present()


    # Slot methods
    def get_selected_slot_ids(self):
        selection = get_tree_selection(self.slot_status_tree)
        ids = map(lambda x: int(self.slot_status_list.get(x[1], 0)[0]),
                  selection)
        return ids


    # Window methods
    def set_restore_icon(self, stock_id):
        icon = gtk.image_new_from_stock(stock_id, gtk.ICON_SIZE_MENU)
        self.restore_tray_item.set_image(icon)


    def get_visible_dialogs(self):
        dialogs = []
        for dialog in self.dialogs:
            if dialog.flags() & gtk.MAPPED:
                dialogs.append(dialog)

        return dialogs
        

    def minimize(self):
        # Hide dialogs
        self.restore_dialogs = self.get_visible_dialogs()
        for dialog in self.restore_dialogs: dialog.hide()

        self.window.hide()
        self.window_visible = False

        widget = self.restore_tray_item.get_child()
        widget.set_text('Show Window')
        widget.set_tooltip_text('Show FAHContol window')
        self.set_restore_icon(gtk.STOCK_GOTO_TOP)


    def restore(self):
        self.window.present()
        self.window_visible = True

        # Restore dialogs
        for dialog in self.restore_dialogs: dialog.present()

        widget = self.restore_tray_item.get_child()
        widget.set_text('Hide Window')
        widget.set_tooltip_text('Hide FAHControl window')
        self.set_restore_icon(gtk.STOCK_GOTO_BOTTOM)


    def toggle(self):
        if self.window_visible: self.minimize()
        else: self.restore()


    # Messages
    def close_error_dialog(self, dialog, id = None, data = None):
        dialog.destroy()
        self.error_dialog = None


    def error(self, message, buttons = gtk.BUTTONS_OK, on_response = None,
              on_response_data = None):
        message = str(message)

        # log to terminal window
        if sys.exc_info()[2]: traceback.print_exc()
        print 'ERROR:', message

        # Don't open more than one
        if self.error_dialog is not None: return False

        if sys.exc_info()[1]: message += '\n%s' % sys.exc_info()[1]

        # create an error message dialog and display modally to the user
        dialog = \
            gtk.MessageDialog(None,
                              gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                              gtk.MESSAGE_ERROR, buttons, message)

        dialog.connect('close', self.close_error_dialog)
        if on_response is not None:
            dialog.connect('response', on_response, on_response_data)
        dialog.connect('response', self.close_error_dialog)
        dialog.set_transient_for(self.window)

        self.error_dialog = dialog
        dialog.show()

        return True


    # Property signals
    def store_property(self, widget, property, name):
        self.db.set(name, widget.get_property(property.name), queue = True)


    def store_dimensions(self, widget, event, name):
        x, y, width, height = widget.get_allocation()
        if 0 <= width and 0 <= height:
            self.db.set(name + '_width', width, queue = True);
            self.db.set(name + '_height', height, queue = True);


    # Systray signals
    def on_restore(self, widget, data = None):
        self.toggle()


    def on_tray_icon_popup_menu(self, widget, button, time, data = None):
        self.tray_menu.popup(None, None, None, button = button,
                             activate_time = time, data = data)
        self.tray_menu.get_parent_window().set_keep_above(True)


    def on_tray_icon_activate(self, widget, data = None):
        self.toggle()
        return True


    def on_tray_menu_enter_notify_event(self, widget, event, data = None):
        self.mouse_in_tray_menu = None


    def on_tray_menu_leave_notify_event(self, widget, event, data = None):
        self.mouse_in_tray_menu = event.time + 1 # Timeout in 1 sec


    # Action signals
    def on_quit(self, widget, data = None):
        self.quit()


    def on_hide(self, widget, data = None):
        self.minimize()


    def on_preferences(self, widget, data = None):
        # OSX crashes with out this, but it's a good idea anyway
        if not self.window_visible: self.restore()

        if self.get_visible_dialogs(): return

        self.preferences_dialog_init()
        self.preferences_dialog.present()


    def on_unpause(self, widget, data = None):
        for client in self.selected_clients:
            client.unpause()


    def on_pause(self, widget, data = None):
        for client in self.selected_clients:
            client.pause()


    def on_finish(self, widget, data = None):
        for client in self.selected_clients:
            client.finish()


    def viewer_check(self):
        if self.viewer is not None and self.viewer.poll() is not None:
            self.viewer = None # Viewer exited


    def viewer_close(self):
        if self.viewer is not None:
            try:
                self.viewer.kill()
                self.viewer.wait() # Note: This could cause a hang kill() fails
            except: pass

            self.viewer = None


    def on_viewer(self, widget, data = None):
        self.viewer_close()

        # Get preferences
        command = self.get_pref('viz_command')
        fullscreen = parse_bool(self.get_pref('viz_fullscreen'))
        width = self.get_pref('viz_width')
        height = self.get_pref('viz_height')
        mode = self.get_viz_render_mode()
        cycle_snapshots = self.get_pref('viz_cycle_snapshots')

        # Create command line
        cmd = shlex.split(command)
        if fullscreen: cmd.append('--fullscreen')
        cmd.append('--width=' + width)
        cmd.append('--height=' + height)
        cmd.append('--mode=%d' % mode)
        cmd.append('--cycle-snapshots=' + cycle_snapshots)
        
        if self.active_client and self.active_client.is_online():
            address = self.active_client.address
            port = self.active_client.port
            cmd.append('--connect=%s:%d' % (address, port))

            slot = self.active_client.get_selected_slot(self)
            if slot is not None: cmd.append('--slot=%d' % slot.id)

        debug = True
        if debug: print cmd

        try:
            self.viewer = subprocess.Popen(cmd, cwd = get_home_dir())
        except Exception, e:
            self.error('Failed to launch viewer with command:\n\n' +
                       ' '.join(cmd))


    def on_mode_changed(self, widget, data = None):
        mode = self.get_mode()
        mode = self.set_mode(mode)
        self.db.set('mode', mode)


    def on_about(self, widget, data = None):
        # OSX crashes with out this, but it's a good idea anyway
        if not self.window_visible: self.restore()

        if self.get_visible_dialogs(): return False

        self.about_dialog.present()


    def on_about_close(self, widget, data = None):
        self.about_dialog.hide()
        return True # Cancel event


    # Window signals
    def on_window_destroy(self, widget, data = None):
        if self.quit_on_close or sys.platform == 'darwin': self.quit()
        else:
            self.minimize()
            return True # Cancel event


    def on_window_delete(self, widget, event, data = None):
        return self.on_window_destroy(widget)


    # Preferences signals
    def on_preferences_ok(self, widget, data = None):
        self.preferences_set()
        self.preferences_save()
        self.preferences_dialog.hide()


    def on_preferences_cancel(self, widget, data = None):
        self.preferences_dialog.hide()
        return True # Cancel event


    def on_theme_pref_changed(self, widget, data = None):
        return # Disabled

        iter = widget.get_active_iter()
        theme = widget.get_model().get_value(iter, 0)
        self.load_theme(theme)


    # Client list signals
    def on_client_add_button_clicked(self, widget, data = None):
        if self.get_visible_dialogs(): return

        # Make client name
        for i in xrange(sys.maxint):
            name = 'client%d' % i
            if not name in self.clients: break

        self.client_entries['name'].set_text(name)

        self.client_dialog.client = None

        for name, widget in self.client_config_tabs.items():
            if name != 'connection': widget.hide()

        self.client_dialog.present()


    def on_client_remove_button_clicked(self, widget, data = None):
        remove_tree_selection(self.client_tree, self.remove_client_path)
        self.save_clients()


    def on_configure(self, widget, data = None):
        if self.get_visible_dialogs(): return

        selection = get_tree_selection(self.client_tree)
        if not len(selection): return
        name = self.client_list.get(selection[0][1], 0)[0]
        client = self.clients[name]
        self.edit_client(client)


    def on_client_tree_view_row_activated(self, widget, path, col, data = None):
        if self.get_visible_dialogs(): return

        iter = self.client_list.get_iter(path)
        name = self.client_list.get(iter, 0)[0]
        client = self.clients[name]
        self.edit_client(client)


    def on_client_selection_changed(self, widget, data = None):
        self.deactivate_client()

        self.selected_clients = self.get_selected_clients()

        # Modify selection
        for client in self.clients.values():
            client.set_selected(client in self.selected_clients)

        self.update_client_status()

        if len(self.clients): self.activate_client()


    # Client options list signals
    def on_client_options_add_button_clicked(self, widget, data = None):
        self.option_present(self.option_list, self.client_dialog)


    def on_client_options_remove_button_clicked(self, widget, data = None):
        remove_tree_selection(self.option_tree)


    # Core options list signals
    def on_core_options_add_button_clicked(self, widget, data = None):
        self.core_option_entry.set_text('')
        self.core_options_dialog.set_transient_for(self.client_dialog)
        self.core_options_dialog.present()


    def on_core_options_remove_button_clicked(self, widget, data = None):
        remove_tree_selection(self.core_option_tree)


    def on_core_options_ok(self, widget, data = None):
        option = self.core_option_entry.get_text().strip()
        if not option:
            self.error('Invalid option')
            return

        self.core_option_list.append([option])
        self.core_options_dialog.hide()


    def on_core_options_cancel(self, widget, data = None):
        self.core_options_dialog.hide()
        return True # Cancel event


    # Client dialog signals
    def on_client_ok(self, widget, data = None):
        name = self.client_entries['name'].get_text()
        address = self.client_entries['address'].get_text()
        port = self.client_entries['port'].get_text()
        password = self.client_entries['password'].get_text()

        if not name:
            self.error('Invalid name')
            return

        if not address:
            self.error('Invalid address')
            return

        if not port:
            self.error('Invalid port')
            return

        port = int(port)

        if self.client_dialog.client:
            client = self.client_dialog.client
            config_hidden = self.client_dialog.config_hidden

            if client.name != name or client.port != port:
                if not config_hidden and client.config.has_changes(self):
                    self.error("Cannot save configuration changes when address "
                               "or port has changed.")
                    return

            if self.update_client(client, name, address, port, password):
                if config_hidden or self.save_client_config(client):
                    self.client_dialog.hide()
        else:
            client = Client(self, name, address, port, password)
            if self.add_client(client): self.client_dialog.hide()


    def on_client_cancel(self, widget, data = None):
        self.client_dialog.hide()
        return True # Cancel event


    def on_local_client_button(self, widget, data = None):
        client = self.clients['local']

        if widget.get_label() == 'Start': client.start(self)
        else: client.stop()


    # User status signals
    def on_team_stats_link_changed(self, widget, data = None):
        self.active_client.config.update_user_info(self)


    # Slot status tree signals
    def on_slot_status_tree_cursor_changed(self, widget, data = None):
        if self.active_client:
            self.active_client.config.select_slot(self)


    # Queue tree signals
    def on_queue_tree_cursor_changed(self, widget, data = None):
        if self.active_client:
            self.active_client.config.select_queue_slot(self)


    # Slot list signals
    def on_slot_add_button_clicked(self, widget, data = None):
        SlotConfig.clear_dialog(self)
        self.slot_dialog.slot_iter = None
        self.slot_dialog.present()


    def on_slot_remove_button_clicked(self, widget, data = None):
        remove_tree_selection(self.slot_tree)


    def on_slot_edit_button_clicked(self, widget, data = None):
        selection = get_tree_selection(self.slot_tree)

        if selection:
            iter = selection[0][1]
            slot = self.slot_list.get(iter, 2)[0].slot
            slot.load_dialog(self)
            self.slot_dialog.slot_iter = iter
            self.slot_dialog.present()


    def on_slot_tree_view_row_activated(self, widget, path, col, data = None):
        iter = self.slot_list.get_iter(path)
        slot = self.slot_list.get(iter, 2)[0].slot
        slot.load_dialog(self)
        self.slot_dialog.slot_iter = iter
        self.slot_dialog.present()


    # Slot dialog signals
    def on_slot_ok(self, widget, data = None):
        if self.slot_dialog.slot_iter is None:
            slot = SlotConfig()
            slot.save_dialog(self)
            slot.add_to_ui(self) # Add to list

        else:
            iter = self.slot_dialog.slot_iter
            id, type, wrapper = self.slot_list.get(iter, 0, 1, 2)
            slot = wrapper.slot
            slot.save_dialog(self)
            if slot.type != type: self.slot_list.set(iter, 1, slot.type)

        self.slot_dialog.hide()


    def on_slot_cancel(self, widget, data = None):
        self.slot_dialog.hide()
        return True # Cancel event


    # Slot options list signals
    def on_slot_options_add_button_clicked(self, widget, data = None):
        self.option_present(self.slot_option_list, self.slot_dialog)


    def on_slot_options_remove_button_clicked(self, widget, data = None):
        remove_tree_selection(self.slot_option_tree)


    # Options edit
    def check_option_name(self, name):
        if not re.match(r'^[a-zA-Z][\w-]*$', name):
            self.error('Invalid option name.')
            return False
        return True


    def on_option_edit(self, cell, path, text, model, column):
        if column == 0 and not self.check_option_name(text): return
        if model[path][column] != text: model[path][column] = text


    def on_core_option_cell_edited(self, cell, path, text, data = None):
        if not text:
            self.error('Invalid option')
            return
        if self.core_option_list[path][0] != text:
            self.core_option_list[path][0] = text


    # Options dialog signals
    def option_present(self, model, parent):
        self.option_name_entry.set_text('')
        self.option_value_entry.set_text('')
        self.options_dialog.option_model = model
        self.options_dialog.set_transient_for(parent)
        self.options_dialog.present()


    def on_options_ok(self, widget, data = None):
        name = self.option_name_entry.get_text().strip()
        value = self.option_value_entry.get_text()

        if not self.check_option_name(name): return

        self.options_dialog.option_model.append([name, value])
        self.options_dialog.hide()


    def on_options_cancel(self, widget, data = None):
        self.options_dialog.hide()
        return True # Cancel event


    # Slot status list signals
    def on_slot_status_tree_view_button_press_event(self, widget, event,
                                                    data = None):
        if event.button != 3: return
        self.slot_menu.popup(None, None, None, button = event.button,
                             activate_time = event.time, data = data)


    def on_unpause_slot_item_activate(self, widget, data = None):
        for id in self.get_selected_slot_ids():
            self.active_client.unpause(id)


    def on_pause_slot_item_activate(self, widget, data = None):
        for id in self.get_selected_slot_ids():
            self.active_client.pause(id)


    def on_finish_slot_item_activate(self, widget, data = None):
        for id in self.get_selected_slot_ids():
            self.active_client.finish(id)


    # Log signals
    def on_download_log_clicked(self, widget, data = None):
        self.active_client.refresh_log()
        self.log.set_text('')


    def on_clear_log_clicked(self, widget, data = None):
        self.log.set_text('')


    # Scale value formatting
    def on_cpu_usage_scale_format_value(self, widget, value, data = None):
        return '%d%%' % value


    def on_checkpoint_scale_format_value(self, widget, value, data = None):
        return '%d min.' % value


    def on_uri_hook(self, widget, url, data = None):
        webbrowser.open(url)
